import unittest
from pronomial import replace_corefs, detect_nouns, score_corefs, word_tokenize


class TestCorefEN(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.lang = "en"
        self.maxDiff = None

    def test_female(self):
        self.assertEqual(
            replace_corefs("Inês said she loves me!"),
            "Inês said Inês loves me !"
        )
        # note the failure to replace "their", plurals are only replaced if
        # plural nouns are detected, multi word references are not caught
        self.assertEqual(
            replace_corefs("Alice invited Marcia to go with her to "
                           "their favorite store"),
            "Alice invited Marcia to go with Alice to their "
            "favorite store"
        )
        self.assertEqual(
            replace_corefs("Bob threatened to kill Alice to make her pay "
                           "her debts"),
            "Bob threatened to kill Alice to make Alice pay Alice debts"
        )
        self.assertEqual(
            replace_corefs(
                "Janet has a husband, Bob, and one son, Sam. "
                "A second child was stillborn in November 2009, causing her "
                "to miss Bristol City's match against Nottingham Forest. "
                "City manager Gary Johnson dedicated their equalising goal "
                "in the match to Janet, who had sent a message of support to her teammates."),
            "Janet has a husband , Bob , and one son , Sam . "
            "A second child was stillborn in November 2009 , causing Janet to miss Bristol City 's match against Nottingham Forest . "
            "City manager Gary Johnson dedicated their equalising goal in the match to Janet , "
            "Janet had sent a message of support to Janet teammates .")

    def test_male(self):
        self.assertEqual(
            replace_corefs("Jack is one of the top candidates in "
                           "the elections. His ideas are unique compared to "
                           "Bob's."),
            "Jack is one of the top candidates in the elections . Jack ideas are unique compared to Bob 's ."
        )
        self.assertEqual(
            replace_corefs("I voted for Elon because he is clear about his "
                           "values. His ideas represent a majority of the "
                           "nation. He is better than Jack."),
            "I voted for Elon because Elon is clear about Elon values ."
            " Elon ideas represent a majority of the nation ."
            " Elon is better than Jack ."
        )
        self.assertEqual(
            replace_corefs(
                "One night, Michael caught Tom breaking into his office to steal drugs, and he used this information to blackmail Lisa into marrying him."),
            "One night , Michael caught Tom breaking into Michael office to steal drugs , and Michael used this information to blackmail Lisa into marrying Michael ."
        )

    def test_plural(self):
        self.assertEqual(
            replace_corefs(
                "Members voted for John because they see him as a good leader."),
            "Members voted for John because Members see John as a good leader ."
        )
        self.assertEqual(
            replace_corefs(
                "Leaders around the world say they stand for peace."),
            "Leaders around the world say Leaders stand for peace ."
        )
        self.assertEqual(
            replace_corefs(
                "I have many friends. They are an important part of my life."),
            "I have many friends . friends are an important part of my life ."
        )
        # they + it
        self.assertEqual(
            replace_corefs(
                "My neighbours just adopted a puppy. They care for it like a baby."),
            "My neighbours just adopted a puppy . neighbours care for puppy like a baby ."
        )

    def test_it(self):
        self.assertEqual(
            replace_corefs("My neighbors have a cat. It has a bushy tail."),
            "My neighbors have a cat . cat has a bushy tail ."
        )
        self.assertEqual(
            replace_corefs("Here is the book now take it."),
            "Here is the book now take book ."
        )
        # dog -> neutral
        self.assertEqual(
            replace_corefs("dog is man's best friend. It is always loyal."),
            "dog is man 's best friend . dog is always loyal ."
        )
        # Dog -> name (male)
        self.assertEqual(
            replace_corefs("Dog is man's best friend. It is always loyal."),
            "Dog is man 's best friend . Dog is always loyal ."
        )
        self.assertEqual(
            replace_corefs("is the light turned on? turn it off"),
            "is the light turned on ? turn light off"
        )
        self.assertEqual(
            replace_corefs("Turn off the light and change it to blue"),
            "Turn off the light and change light to blue"
        )

        # it + who
        self.assertEqual(
            replace_corefs(
                "London is the capital and most populous city of England and the United Kingdom. "
                "Standing on the River Thames in the south east of the island of Great Britain, "
                "London has been a major settlement for two millennia. "
                "It was founded by the Romans, who named it Londinium."),
            "London is the capital and most populous city of England and the United Kingdom . "
            "Standing on the River Thames in the south east of the island of Great Britain , "
            "London has been a major settlement for two millennia . "
            "London was founded by the Romans , Romans named London Londinium ."
        )

    def test_dictionary_words(self):
        # harcoded wordlists that map words to gendered pronouns
        # boy -> male
        self.assertEqual(
            replace_corefs(
                "The sign was too far away for the boy to read it."),
            "The sign was too far away for the boy to read sign ."
        )
        # girl -> female
        self.assertEqual(
            replace_corefs("The girl said she would take the trash out."),
            "The girl said girl would take the trash out ."
        )
        # mom -> female
        self.assertEqual(
            replace_corefs("call Mom. tell her to buy eggs. tell her to buy "
                           "coffee. tell her to buy milk"),
            "call Mom . tell Mom to buy eggs . tell Mom to buy coffee . tell Mom to buy milk"
        )
        # dad -> male
        self.assertEqual(
            replace_corefs("call dad. tell him to buy bacon. tell him to buy "
                           "coffee. tell him to buy beer"),
            "call dad . tell dad to buy bacon . tell dad to buy coffee . tell dad to buy beer"
        )

    def test_hardcoded_postag_fixes(self):
        # "turn on" falsely detected as ('Turn', 'NN'), ('on', 'IN')
        # "change it" falsely detected as ('change', 'NN'), ('it', 'PRP')
        # these specific cases are corrected manually when detected
        self.assertEqual(
            replace_corefs("Turn on the light and change it to blue"),
            "Turn on the light and change light to blue"
        )

    def test_mixed(self):
        self.assertEqual(
            replace_corefs("My sister has a dog, She loves him!"),
            "My sister has a dog , sister loves dog !"
        )
        self.assertEqual(
            replace_corefs(
                "A short while later, Michael decided that he wanted to play a role in his son's life, and tried to get Lisa to marry him, but by this time, she wanted nothing to do with him. "
                "Around the same time, Lisa's son Tom had returned from Vietnam with a drug habit. One night, Michael caught Tom breaking into his office to steal drugs"
            ),
            "A short while later , Michael decided that Michael wanted to play a role in Michael son 's life , and tried to get Lisa to marry Michael , but by this time , Lisa wanted nothing to do with Michael . "
            "Around the same time , Lisa 's son Tom had returned from Vietnam with a drug habit . One night , Michael caught Tom breaking into Michael office to steal drugs"
        )

    def test_neutral_postag(self):
        # in {noun}-> in == IN -> {noun} == neutral
        # "Arendelle" is recognized as a location because of "in", otherwise
        # it would be tagged as a female name and be used instead of Elsa
        self.assertEqual(
            replace_corefs(
                "Chris is very handsome. He is Australian. Elsa lives in Arendelle. He likes her."),
            "Chris is very handsome . Chris is Australian . Elsa lives in Arendelle . Chris likes Elsa ."
        )

    def test_in(self):
        # "into/in his/her/their" will select first male/female/plural noun
        # instead of last
        self.assertEqual(
            replace_corefs("One night, Michael caught Tom in his office"),
            "One night , Michael caught Tom in Michael office"
        )
        self.assertEqual(
            replace_corefs(
                "One night, Michael caught Tom breaking into his office"),
            "One night , Michael caught Tom breaking into Michael office"
        )

    def test_with(self):
        # "with her/him/them" will select first male/female/plural noun
        # instead of last seen
        self.assertEqual(
            replace_corefs("Alice invited Marcia to go with her"),
            "Alice invited Marcia to go with Alice"
        )
        self.assertEqual(
            replace_corefs("The Martians invited the Venusians to go with "
                           "them to Pluto"),
            "The Martians invited the Venusians to go with Martians to Pluto"
        )
        self.assertEqual(
            replace_corefs("Kevin invited Bob to go with him to "
                           "his favorite fishing spot"),
            "Kevin invited Bob to go with Kevin to Bob favorite fishing spot"
        )

    def test_that(self):
        # "that her/he/they" will select first male/female/plural noun
        # instead of last seen
        self.assertEqual(
            replace_corefs(
                "Bob telephoned Jake to tell him that he lost the laptop."),
            "Bob telephoned Jake to tell Jake that Bob lost the laptop ."
        )
        self.assertEqual(
            replace_corefs("Ana telephoned Alice to tell her that she lost "
                           "the bus"),
            "Ana telephoned Alice to tell Alice that Ana lost the bus"
        )
        self.assertEqual(
            replace_corefs("The Martians told the Venusians that they used "
                           "to have an ocean"),
            "The Martians told the Venusians that Martians used to have an ocean"
        )

    def test_capitalization_ambiguity(self):
        # capitalization mistakes
        # Cat with upper case is tagged as a male name, works as expected
        self.assertEqual(
            replace_corefs(
                "My neighbors have a Cat. It has a bushy tail. They love him!"),
            "My neighbors have a Cat . Cat has a bushy tail . neighbors love Cat !"
        )

        # cat with lower case is not gender classified and not selected to
        # replace "him", tail is wrongly selected because it is the closest
        # noun
        self.assertEqual(
            replace_corefs("My neighbors have a cat. It has a bushy tail. "
                           "They love him!"),
            "My neighbors have a cat . cat has a bushy tail . neighbors love tail !"
            # ... neighbors love cat ...
        )

    def test_ambiguous(self):
        # a little ambiguous results, unexpected but technically acceptable
        self.assertEqual(
            replace_corefs("Joe was talking to Bob and told him to go home "
                           "because he was drunk"),
            "Joe was talking to Bob and told Bob to go home because Joe was drunk"
            # ... because Bob was drunk ...
        )

        self.assertEqual(
            replace_corefs("the dudes were talking with their enemies and "
                           "they decided to avoid war"),
            "the dudes were talking with dudes enemies and enemies decided to avoid war"
            # ... and dudes decided ...
        )

    def test_gender_failures(self):
        ## gender misclassifications
        # NOTE Sproule is misclassified as female
        self.assertEqual(
            replace_corefs(
                "Janet has a husband, Sproule, and one son, Sam. "
                "A second child was stillborn in November 2009, causing her "
                "to miss Bristol City's match against Nottingham Forest. "
                "City manager Gary Johnson dedicated their equalising goal "
                "in the match to Janet, who had sent a message of support to her teammates."),
            "Janet has a husband , Sproule , and one son , Sam . "
            "A second child was stillborn in November 2009 , causing Janet to miss Bristol City 's match against Nottingham Forest . "
            "City manager Gary Johnson dedicated their equalising goal in the match to Janet , "
            "Janet had sent a message of support to Janet teammates .")

        # "his" and "him" are not properly replaced because Sproule is misclassified as female
        self.assertEqual(
            replace_corefs(
                "Sproule has a wife, Janet, and one son, Sam. "
                "A second child was stillborn in November 2009, causing him to miss Bristol City's match against Nottingham Forest. "
                "City manager Gary Johnson dedicated their equalising goal in the match to Sproule, who had sent a message of support to his teammates."),
            "Sproule has a wife , Janet , and one son , Sam . "
            "A second child was stillborn in November 2009 , causing Sam to miss Bristol City 's match against Nottingham Forest . "
            "City manager Gary Johnson dedicated their equalising goal in the match to Sproule , Sproule had sent a message of support to Johnson teammates .")
        #  "... causing Sproule to miss ... to Sproule teammates"

    def test_detect_nouns(self):
        tokens = ['London', 'has', 'been', 'a', 'major', 'settlement', 'for',
                  'two', 'millennia', '.', 'It', 'was', 'founded', 'by', 'the',
                  'Romans', ',', 'who', 'named', 'it', 'Londinium', '.']
        self.assertEqual(
            detect_nouns(tokens, lang="en"),
            {'female': [],
             'first': [],
             'male': [0, 20],
             'neutral': [0, 5, 8, 20],
             'plural': [15],
             'subject': [0, 5, 8, 20],
             'verb_subject': [0],
             "tokens": tokens})
        self.assertEqual(
            detect_nouns(tokens, lang="en", return_idx=False),
            {'female': [],
             'first': [],
             'male': ['London', 'Londinium'],
             'neutral': ['London', 'settlement', 'millennia', 'Londinium'],
             'plural': ['Romans'],
             'subject': ['London', 'settlement', 'millennia', 'Londinium'],
             'verb_subject': ['London'],
             "tokens": tokens})

    def test_coreferences(self):

        # word indexes are used internally, but to make these tests human
        # friendly we are comparing actual words, note that in practice we
        # should work with token indexes
        def test_prediction(sentence, expected):
            tokens = word_tokenize(sentence)
            pred = score_corefs(sentence, lang="en")
            matches = []
            for tok_idx, match in pred.items():
                tok = tokens[tok_idx]
                for tok2_idx, score in match.items():
                    tok2 = tokens[tok2_idx]
                    matches.append((tok, tok2, score))
            self.assertEqual(matches, expected)

        test_prediction(
            "London has been a major settlement for two millennia. It was founded by the Romans, who named it Londinium.",
            [('It', 'London', 0.56),
             ('It', 'settlement', 0.06),
             ('It', 'millennia', 0.39),
             ('who', 'settlement', 0.02),
             ('who', 'millennia', 0.28),
             ('who', 'Romans', 0.7),
             ('it', 'London', 0.56),
             ('it', 'settlement', 0.06),
             ('it', 'millennia', 0.39)]
            )
        test_prediction("My neighbors have a Cat. It has a bushy tail. They "
                        "love him!",
                        [('It', 'Cat', 1.0),
                         ('They', 'neighbors', 1.0),
                         ('him', 'Cat', 0.65),
                         ('him', 'tail', 0.35)]
                        )
        test_prediction("dog is man's best friend. It is always loyal.",
                        [('It', 'dog', 0.56),
                         ('It', 'man', 0.06),
                         ('It', 'friend', 0.39)]
                        )
        test_prediction("The sign was too far away for the boy to read it.",
                        [('it', 'sign', 0.62),
                         ('it', 'boy', 0.38)]
                        )
        test_prediction("call dad. tell him to buy bacon. tell him to buy "
                        "coffee. tell him to buy beer",
                        [('him', 'dad', 1.0),
                         ('him', 'dad', 0.64),
                         ('him', 'bacon', 0.36),
                         ('him', 'dad', 0.58),
                         ('him', 'bacon', 0.06),
                         ('him', 'coffee', 0.36)]
                        )
        test_prediction(
            "the dudes were talking with their enemies and they decided to avoid war",
            [('their', 'dudes', 1.0),
             ('they', 'dudes', 0.33),
             ('they', 'enemies', 0.67)]
            )

        test_prediction(
            "Alice invited Marcia to go with her to their favorite store",
            [('her', 'Alice', 0.68),
             ('her', 'Marcia', 0.32)]
        )

        test_prediction(
            "Joe was talking to Bob and told him to go home because he was drunk",
            [('him', 'Joe', 0.47),
             ('him', 'Bob', 0.53),
             ('he', 'Joe', 0.51),
             ('he', 'Bob', 0.35),
             ('he', 'home', 0.14)]
        )

        test_prediction(
            "One night, Michael caught Tom breaking into his office to steal drugs, and he used this information to blackmail Lisa into marrying him.",
            [('his', 'Michael', 0.57),
             ('his', 'Tom', 0.43),
             ('he', 'Michael', 0.51),
             ('he', 'Tom', 0.35),
             ('he', 'office', 0.14),
             ('him', 'Michael', 0.46),
             ('him', 'Tom', 0.32),
             ('him', 'office', 0.03),
             ('him', 'information', 0.04),
             ('him', 'Lisa', 0.15)]
        )
        test_prediction(
            "A short while later, Michael decided that he wanted to play a role in his son's life, and tried to get Lisa to marry him, but by this time, she wanted nothing to do with him. "
            "Around the same time, Lisa's son Tom had returned from Vietnam with a drug habit. One night, Michael caught Tom breaking into his office to steal drugs",
            [('he', 'while', 0.19),
             ('he', 'Michael', 0.81),
             ('his', 'while', 0.02),
             ('his', 'Michael', 0.93),
             ('his', 'role', 0.05),
             ('him', 'while', 0.01),
             ('him', 'Michael', 0.62),
             ('him', 'role', 0.04),
             ('him', 'son', 0.05),
             ('him', 'life', 0.07),
             ('him', 'Lisa', 0.21),
             ('she', 'while', 0.02),
             ('she', 'Michael', 0.03),
             ('she', 'role', 0.05),
             ('she', 'son', 0.07),
             ('she', 'life', 0.09),
             ('she', 'Lisa', 0.45),
             ('she', 'time', 0.29),
             ('him', 'while', 0.01),
             ('him', 'Michael', 0.52),
             ('him', 'role', 0.03),
             ('him', 'son', 0.04),
             ('him', 'life', 0.05),
             ('him', 'Lisa', 0.07),
             ('him', 'time', 0.19),
             ('him', 'nothing', 0.09),
             ('his', 'Michael', 0.1),
             ('his', 'role', 0.01),
             ('his', 'son', 0.01),
             ('his', 'life', 0.01),
             ('his', 'Lisa', 0.01),
             ('his', 'time', 0.01),
             ('his', 'nothing', 0.02),
             ('his', 'time', 0.02),
             ('his', 'Lisa', 0.02),
             ('his', 'son', 0.02),
             ('his', 'Tom', 0.14),
             ('his', 'Vietnam', 0.11),
             ('his', 'drug', 0.03),
             ('his', 'habit', 0.03),
             ('his', 'night', 0.03),
             ('his', 'Michael', 0.21),
             ('his', 'Tom', 0.21)]
        )






from pronomial.utils import predict_gender, pos_tag, word_tokenize, is_plural


class PronomialCoreferenceSolver:
    @staticmethod
    def _load_lang_resources(lang):
        # universal tagset, depends on pos tagger model for each lang
        PRONOUN_TAG = ['PRON']
        NOUN_TAG = ['NOUN']
        JJ_TAG = ['ADJ']
        PLURAL_NOUN_TAG = ['NOUN']
        SUBJ_TAG = ['NOUN']

        # this is a basic mechanism to allow to "look back", the naming
        # reflects what this does poorly, each language can define word
        # pairs that will give preference to an earlier noun in the
        # sentence, loosely named after a couple examples in the
        # english unittests. TODO generalize this
        WITH = WITH_FOLLOWUP = THAT = THAT_FOLLOWUP = IN = IN_FOLLOWUP = []

        NEUTRAL_WORDS = []

        # symbol used when merging Nouns to replace plural pronouns
        NAME_JOINER = "+"

        # word lists
        PRONOUNS = {}
        GENDERED_WORDS = {}

        # Load resources for selected lang
        if lang.startswith("en"):
            from pronomial.lang.en import NOUN_TAG_EN, PLURAL_NOUN_TAG_EN, \
                PRONOUNS_EN, PRONOUN_TAG_EN, SUBJ_TAG_EN, JJ_TAG_EN, WITH_EN, \
                GENDERED_WORDS_EN, WITH_FOLLOWUP_EN, THAT_EN, \
                THAT_FOLLOWUP_EN, NEUTRAL_WORDS_EN, NAME_JOINER_EN, IN_EN, \
                IN_FOLLOWUP_EN
            GENDERED_WORDS = GENDERED_WORDS_EN
            NOUN_TAG = NOUN_TAG_EN
            SUBJ_TAG = SUBJ_TAG_EN
            PRONOUN_TAG = PRONOUN_TAG_EN
            PRONOUNS = PRONOUNS_EN
            PLURAL_NOUN_TAG = PLURAL_NOUN_TAG_EN
            JJ_TAG = JJ_TAG_EN
            WITH = WITH_EN
            WITH_FOLLOWUP = WITH_FOLLOWUP_EN
            THAT = THAT_EN
            THAT_FOLLOWUP = THAT_FOLLOWUP_EN
            NEUTRAL_WORDS = NEUTRAL_WORDS_EN
            NAME_JOINER = NAME_JOINER_EN
            IN = IN_EN
            IN_FOLLOWUP = IN_FOLLOWUP_EN
        elif lang.startswith("pt"):
            from pronomial.lang.pt import PRONOUNS_PT, GENDERED_WORDS_PT
            GENDERED_WORDS = GENDERED_WORDS_PT
            PRONOUNS = PRONOUNS_PT
        elif lang.startswith("es"):
            from pronomial.lang.es import PRONOUNS_ES, GENDERED_WORDS_ES
            GENDERED_WORDS = GENDERED_WORDS_ES
            PRONOUNS = PRONOUNS_ES
        elif lang.startswith("ca"):
            from pronomial.lang.ca import PRONOUNS_CA, GENDERED_WORDS_CA
            GENDERED_WORDS = GENDERED_WORDS_CA
            PRONOUNS = PRONOUNS_CA

        return PRONOUN_TAG, NOUN_TAG, JJ_TAG, PLURAL_NOUN_TAG, SUBJ_TAG, \
               WITH, WITH_FOLLOWUP, THAT, THAT_FOLLOWUP, IN, IN_FOLLOWUP, \
               NEUTRAL_WORDS, NAME_JOINER, PRONOUNS, \
               GENDERED_WORDS

    @staticmethod
    def detect_nouns(sentence, lang="en", return_idx=True):

        PRONOUN_TAG, NOUN_TAG, JJ_TAG, PLURAL_NOUN_TAG, SUBJ_TAG, \
        WITH, WITH_FOLLOWUP, THAT, THAT_FOLLOWUP, IN, IN_FOLLOWUP, \
        NEUTRAL_WORDS, NAME_JOINER, PRONOUNS, GENDERED_WORDS = \
            PronomialCoreferenceSolver._load_lang_resources(lang)

        tags = pos_tag(sentence, lang=lang)
        prev_names_idx = {
            "male": [],
            "female": [],
            "first": [],
            "neutral": [],
            "plural": [],
            "subject": [],
            "verb_subject": []
        }

        for idx, (w, t) in enumerate(tags):
            next_w, next_t = tags[idx + 1] if idx < len(tags) - 1 else ("", "")
            prev_w, prev_t = tags[idx - 1] if idx > 0 else ("", "")
            if t in NOUN_TAG:
                if prev_w in NEUTRAL_WORDS:
                    prev_names_idx["neutral"].append(idx)
                else:
                    gender = predict_gender(w, prev_w, lang=lang)
                    if w in PRONOUNS["female"] or \
                            w.lower() in GENDERED_WORDS["female"]:
                        prev_names_idx["female"].append(idx)
                    elif w in PRONOUNS["male"] or \
                            w.lower() in GENDERED_WORDS["male"]:
                        prev_names_idx["male"].append(idx)
                    elif w[0].isupper() or prev_t in ["DET"]:
                        prev_names_idx[gender].append(idx)
                    prev_names_idx["subject"].append(idx)
                    prev_names_idx["neutral"].append(idx)

                if next_t.startswith("V") and not prev_t.startswith("V"):
                    prev_names_idx["verb_subject"].append(idx)

            elif t in SUBJ_TAG:
                prev_names_idx["subject"].append(idx)
                gender = predict_gender(w, prev_w, lang=lang)
                prev_names_idx["neutral"].append(idx)
                if gender == "female":
                    prev_names_idx["female"].append(idx)
                if gender == "male":
                    prev_names_idx["male"].append(idx)
            elif t in PLURAL_NOUN_TAG:
                prev_names_idx["plural"].append(idx)
                if w[0].isupper():
                    gender = predict_gender(w, prev_w, lang=lang)
                    if not prev_names_idx[gender]:
                        prev_names_idx[gender].append(idx)
            elif t in JJ_TAG:  # common tagger error
                if w[0].isupper():
                    gender = predict_gender(w, prev_w, lang=lang)
                    if not prev_names_idx[gender]:
                        prev_names_idx[gender].append(idx)
                if not prev_names_idx["neutral"]:
                    prev_names_idx["neutral"].append(idx)
                if not prev_names_idx["subject"]:
                    prev_names_idx["subject"].append(idx)

        if isinstance(sentence, str):
            tokens = word_tokenize(sentence)
        else:
            tokens = sentence

        if not return_idx:
            prev_names_idx = {k: [tokens[i] for i in v]
                              for k, v in prev_names_idx.items()}
        prev_names_idx["tokens"] = tokens
        return prev_names_idx

    @staticmethod
    def score_corefs(sentence, lang="en"):
        PRONOUN_TAG, NOUN_TAG, JJ_TAG, PLURAL_NOUN_TAG, SUBJ_TAG, \
        WITH, WITH_FOLLOWUP, THAT, THAT_FOLLOWUP, IN, IN_FOLLOWUP, \
        NEUTRAL_WORDS, NAME_JOINER, PRONOUNS, \
        GENDERED_WORDS = PronomialCoreferenceSolver._load_lang_resources(lang)

        tags = pos_tag(sentence, lang=lang)
        pron_list = [p for k, p in PRONOUNS.items()]
        flatten = lambda l: [item for sublist in l for item in sublist]
        pron_list = flatten(pron_list)

        tagged_nouns = PronomialCoreferenceSolver.detect_nouns(sentence, lang)
        tokens = tagged_nouns.pop("tokens")
        candidates = {}

        for idx, (w, t) in enumerate(tags):
            if (t in PRONOUN_TAG and w.lower() in pron_list) or \
                    any(w in items for k, items in PRONOUNS.items()):
                candidates[idx] = {}

                prev_w, prev_t = tags[idx - 1] if idx > 0 else ("", "")
                prev_ids = {k: [i for i in v if i <= idx]
                            for k, v in tagged_nouns.items()}
                prev_names = {k: [tokens[i] for i in v]
                              for k, v in prev_ids.items()}
                # this is a basic mechanism to allow to "look back", the naming
                # reflects what this does poorly, each language can define word
                # pairs that will give preference to an earlier noun in the
                # sentence, loosely named after a couple examples in the
                # english unittests. TODO generalize this
                idz = -1
                if prev_w.lower() in WITH and w.lower() in WITH_FOLLOWUP:
                    idz = -2
                elif prev_w.lower() in THAT and w.lower() in THAT_FOLLOWUP:
                    idz = -2
                elif prev_w.lower() in IN and w.lower() in IN_FOLLOWUP:
                    idz = -2
                wl = w.lower()

                # score subject tags
                if wl in PRONOUNS["neutral"] or \
                        wl in PRONOUNS["male"] or \
                        wl in PRONOUNS["female"] or \
                        (wl in PRONOUNS["plural"] and t in ["WP"]):  # "who"

                    if prev_names["subject"]:
                        if idz < 0:
                            idz = len(prev_names["subject"]) + idz
                        for ids, x in enumerate(prev_ids["subject"]):
                            if x not in candidates[idx]:
                                candidates[idx][x] = 0
                            if ids == idz:
                                candidates[idx][x] += 10  # freshness bonus
                            candidates[idx][x] += ids

                # score gendered pronoun tags
                if wl in PRONOUNS["male"]:
                    if prev_names["male"]:
                        if idz < 0:
                            idz = len(prev_names["male"]) + idz
                        for ids, x in enumerate(prev_ids["male"]):
                            if x not in candidates[idx]:
                                candidates[idx][x] = 0
                            candidates[idx][x] += 20  # gender match
                            # freshness bonus
                            candidates[idx][x] += ids * 10
                            if ids == idz:
                                candidates[idx][x] += 10
                            if tokens[x] in [_ for _ in prev_names["male"]
                                             if _ in prev_names["verb_subject"]]:
                                candidates[idx][x] += 25  # verb subject match

                elif wl in PRONOUNS["female"]:
                    if prev_names["female"]:
                        if idz < 0:
                            idz = len(prev_names["female"]) + idz
                        for ids, x in enumerate(prev_ids["female"]):
                            if x not in candidates[idx]:
                                candidates[idx][x] = 0

                            candidates[idx][x] += 20  # gender match
                            # freshness bonus
                            candidates[idx][x] += ids * 10
                            if ids == idz:
                                candidates[idx][x] += 10
                            if tokens[x] in [_ for _ in prev_names["female"]
                                             if _ in prev_names["verb_subject"]]:
                                candidates[idx][x] += 25  # verb subject match

                elif wl in PRONOUNS["neutral"]:
                    if prev_names["neutral"]:
                        for ids, x in enumerate(prev_ids["neutral"]):
                            if x not in candidates[idx]:
                                candidates[idx][x] = 0
                            # freshness bonus
                            candidates[idx][x] += ids

                            if tokens[x] in prev_names["verb_subject"]:
                                candidates[idx][x] += 20  # verb subject match

                # score plural tags
                elif wl in PRONOUNS["plural"]:
                    if prev_names["plural"]:
                        if idz < 0:
                            idz = len(prev_names["plural"]) + idz
                        for ids, x in enumerate(prev_ids["plural"]):
                            if x not in candidates[idx]:
                                candidates[idx][x] = 0
                            candidates[idx][x] += 20  # plural pronoun match
                            # freshness bonus
                            candidates[idx][x] += ids * 10
                            if ids == idz:
                                candidates[idx][x] += 10
                            if tokens[x] in [_ for _ in prev_names["subject"] if
                                             is_plural(_, lang)]:
                                candidates[idx][x] += 15  # plural word match
                            if ids in prev_ids["verb_subject"]:
                                candidates[idx][x] += 10  # verb subject match

                # catch all scorer
                # nothing matched but we have a pronoun, if it's not
                # a first person pronoun let's score all subject tags
                elif prev_names["subject"] and wl not in PRONOUNS["first"]:
                    if idz < 0:
                        idz = len(prev_names["subject"]) + idz
                    for ids, x in enumerate(prev_ids["subject"]):
                        if x not in candidates[idx]:
                            candidates[idx][x] = 0
                        if ids == idz:
                            candidates[idx][x] += 10  # freshness bonus
                        candidates[idx][x] += ids

        # make score of all results add up to 1
        for tok_idx in candidates:
            total = sum(
                score for tok2_idx, score in candidates[tok_idx].items())
            for tok2_idx, score in candidates[tok_idx].items():
                candidates[tok_idx][tok2_idx] = round(score / total, 2)
        return {k: {k2: v2 for k2, v2 in v.items() if v2 > 0}
                for k, v in candidates.items()}

    @staticmethod
    def solve_corefs(sentence, lang="en", return_idx=True):
        tokens = word_tokenize(sentence)
        candidates = PronomialCoreferenceSolver.score_corefs(sentence, lang)
        corefs = []
        for tok_id, match in candidates.items():
            matches = [(tok_id, mtok_id, score)
                       for mtok_id, score in match.items()]
            matches = sorted(matches, key=lambda k: k[2], reverse=True)
            if len(matches):
                corefs.append(matches[0])
        if return_idx:
            return corefs
        return [(tokens[tok_id], tokens[mtok_id], score)
                for tok_id, mtok_id, score in corefs]

    @classmethod
    def replace_corefs(cls, text, lang="en"):
        tokens = word_tokenize(text)
        for tok_id, mtok_id, score in cls.solve_corefs(text, lang=lang):
            tokens[tok_id] = tokens[mtok_id]
        return " ".join(tokens)

    @staticmethod
    def normalize(text):
        return " ".join(word_tokenize(text))


def normalize(*args, **kwargs):
    return PronomialCoreferenceSolver.normalize(*args, **kwargs)


def detect_nouns(*args, **kwargs):
    return PronomialCoreferenceSolver.detect_nouns(*args, **kwargs)


def link_pronouns(*args, **kwargs):
    return PronomialCoreferenceSolver.solve_corefs(*args, **kwargs)


def score_corefs(*args, **kwargs):
    return PronomialCoreferenceSolver.score_corefs(*args, **kwargs)


def replace_corefs(*args, **kwargs):
    return PronomialCoreferenceSolver.replace_corefs(*args, **kwargs)




from os.path import join, dirname, isfile
import csv
import nltk
import pickle

from pronomial.lang.pt import pos_tag_pt
from pronomial.lang.en import pos_tag_en, is_plural_en
from pronomial.lang.es import pos_tag_es
from pronomial.lang.ca import pos_tag_ca

from nltk.tokenize import word_tokenize as _wt


def word_tokenize(text, lang="en"):
    try:
        return _wt(text)
    except LookupError:
        nltk.download("punkt")
        return _wt(text)


def pos_tag(text, lang="en"):
    if lang.startswith("en"):
        return pos_tag_en(text)
    if lang.startswith("pt"):
        return pos_tag_pt(text)
    if lang.startswith("es"):
        return pos_tag_es(text)
    if lang.startswith("pt"):
        return pos_tag_ca(text)
    raise NotImplementedError


def _get_features(name):
    return {
        'suffix1': name[-1],
        'suffix2': name[-2:],
        'suffix3': name[-3:],
        'suffix4': name[-4:],
        'suffix5': name[-5:]
    }


def train_gender_classifier(path):
    dataset = join(dirname(__file__), "res", "names.csv")
    with open(dataset) as f:
        names = [tuple(line) for line in csv.reader(f)]
    train_data = [(_get_features(n), g) for (n, g) in names]
    clf = nltk.NaiveBayesClassifier.train(train_data)
    path = path or join(dirname(__file__), "res", "name_gender.pkl")
    with open(path, "wb") as f:
        pickle.dump(clf, f)
    return clf


def load_gender_classifier(path=None):
    path = path or join(dirname(__file__), "res", "name_gender.pkl")
    if not isfile(path):
        train_gender_classifier(path)
    with open(path, "rb") as f:
        tagger = pickle.load(f)
    return tagger


GENDER = load_gender_classifier()


def predict_gender(word, text="", lang="en"):
    gender = None
    if lang.startswith("en"):
        from pronomial.lang.en import GENDERED_WORDS_EN
        for k, v in GENDERED_WORDS_EN.items():
            if word.lower() in v:
                gender = k
                break
    if lang.startswith("pt"):
        from pronomial.lang.pt import predict_gender_pt
        gender = predict_gender_pt(word, text)
    if lang.startswith("es"):
        from pronomial.lang.es import predict_gender_es
        gender = predict_gender_es(word, text)
    if lang.startswith("ca"):
        from pronomial.lang.ca import predict_gender_ca
        gender = predict_gender_ca(word, text)

    return gender or GENDER.classify(_get_features(word))


def is_plural(text, lang="en"):
    if lang.startswith("en"):
        return is_plural_en(text)
    return text.endswith("s")



import unittest
from pronomial import replace_corefs, word_tokenize, score_corefs


class TestCorefPT(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.lang = "pt"

    def test_female(self):
        # female only
        self.assertEqual(
            replace_corefs("A Ana gosta de cães. Ela tem dois",
                           lang=self.lang),
            "A Ana gosta de cães . Ana tem dois"
        )
        # female x2
        self.assertEqual(
            replace_corefs("a Ana disse á Maria que ela tem bom gosto",
                           lang=self.lang),
            "a Ana disse á Maria que Maria tem bom gosto"
        )
        # male + female
        self.assertEqual(
            replace_corefs("o João disse ao Joaquim que gosta da Ana. Ela "
                           "é bonita",
                           lang=self.lang),
            "o João disse ao Joaquim que gosta da Ana . Ana é bonita"
        )

    def test_male(self):
        # male only
        self.assertEqual(
            replace_corefs("o João gosta de gatos. Ele tem quatro",
                           lang=self.lang),
            "o João gosta de gatos . João tem quatro"
        )
        # male x2
        self.assertEqual(
            replace_corefs("o João disse ao Joaquim que ele está gordo",
                           lang=self.lang),
            "o João disse ao Joaquim que Joaquim está gordo"
        )

    def test_plural(self):
        # female noun from wordlist
        self.assertEqual(
            replace_corefs("As mulheres da ribeira do sado é que é, "
                           "Elas lavram a terra com as unhas dos pés",
                           lang=self.lang),
            "As mulheres da ribeira do sado é que é , mulheres lavram a terra com as unhas dos pés"
        )
        # male noun from wordlist
        self.assertEqual(
            replace_corefs("Os americanos foram á lua. Eles são fodidos",
                           lang=self.lang),
            "Os americanos foram á lua . americanos são fodidos"
        )

    def test_coreferences(self):

        # word indexes are used internally, but to make these tests human
        # friendly we are comparing actual words, note that in practice we
        # should work with token indexes
        def test_prediction(sentence, expected):
            tokens = word_tokenize(sentence)
            pred = score_corefs(sentence, lang="pt")
            matches = []
            for tok_idx, match in pred.items():
                tok = tokens[tok_idx]
                for tok2_idx, score in match.items():
                    tok2 = tokens[tok2_idx]
                    matches.append((tok, tok2, score))
            self.assertEqual(matches, expected)

        test_prediction(
            "o João disse ao Joaquim que ele está gordo",
            [('ele', 'João', 0.47),
             ('ele', 'Joaquim', 0.53)]
        )

        test_prediction(
            "a Ana disse á Maria que ela tem bom gosto",
            [('ela', 'Ana', 0.47),
             ('ela', 'Maria', 0.53)]
        )


import pickle
from os.path import isfile, join, dirname
from pronomial.utils import word_tokenize
import nltk.tag
import nltk
from string import punctuation
from random import shuffle


PRONOUNS_PT = {
    'male': ['ele', "lo", "dele", "nele", "seu", "eles", "seus", "deles",
             "neles"],
    'female': ['ela', "la", "dela", "nela", "sua", "elas", "suas", "delas",
               "nelas"],
    'first': ['eu', 'me', 'mim', 'nós', "comigo", "meu", "minha", "meus",
              "minhas"],
    'neutral': ["tu", "te", "ti", "lhe", "contigo", "consigo", "si"],
    'plural': ['eles', 'elas', "vós", "vocês", "lhes", "los", "las",
               "neles", "nelas", "convosco", "conosco", "connosco", "teus",
               "tuas", "seus", "suas", "nossos", "vossos", "nossas", "vossas"]
}

# special cases, word2gender mappings
GENDERED_WORDS_PT = {
    "female": ["mãe", "irmã", "tia", "amiga", "prima", "namorada", "mulher",
               "mulheres", "rapariga", "raparigas", "gaja", "gajas", "moça",
               "moças", "elas", "suas"],
    "male": ["pai", "irmão", "tio", "amigo", "primo", "namorado", "homem",
             "homens", "rapaz", "rapazes", "gajo", "gajos", "moço", "moços",
             "eles", "seus"]
}

# context rules for gender
MALE_DETERMINANTS_PT = ["o", "os", "este", "estes", "esse", "esses"]
FEMALE_DETERMINANTS_PT = ["a", "as", "estas", "estas", "essa", "essas"]


def train_pt_tagger(path):
    nltk.download('mac_morpho')
    nltk.download('floresta')

    def convert_to_universal_tag(t, reverse=False):
        tagdict = {
            'n': "NOUN",
            'num': "NUM",
            'v-fin': "VERB",
            'v-inf': "VERB",
            'v-ger': "VERB",
            'v-pcp': "VERB",
            'pron-det': "PRON",
            'pron-indp': "PRON",
            'pron-pers': "PRON",
            'art': "DET",
            'adv': "ADV",
            'conj-s': "CONJ",
            'conj-c': "CONJ",
            'conj-p': "CONJ",
            'adj': "ADJ",
            'ec': "PRT",
            'pp': "ADP",
            'prp': "ADP",
            'prop': "NOUN",
            'pro-ks-rel': "PRON",
            'proadj': "PRON",
            'prep': "ADP",
            'nprop': "NOUN",
            'vaux': "VERB",
            'propess': "PRON",
            'v': "VERB",
            'vp': "VERB",
            'in': "X",
            'prp-': "ADP",
            'adv-ks': "ADV",
            'dad': "NUM",
            'prosub': "PRON",
            'tel': "NUM",
            'ap': "NUM",
            'est': "NOUN",
            'cur': "X",
            'pcp': "VERB",
            'pro-ks': "PRON",
            'hor': "NUM",
            'pden': "ADV",
            'dat': "NUM",
            'kc': "ADP",
            'ks': "ADP",
            'adv-ks-rel': "ADV",
            'npro': "NOUN",
        }
        if t in ["N|AP", "N|DAD", "N|DAT", "N|HOR", "N|TEL"]:
            t = "NUM"
        if reverse:
            if "|" in t: t = t.split("|")[0]
        else:
            if "+" in t: t = t.split("+")[1]
            if "|" in t: t = t.split("|")[1]
            if "#" in t: t = t.split("#")[0]
        t = t.lower()
        return tagdict.get(t, "." if all(tt in punctuation for tt in t) else t)

    floresta = [[(w, convert_to_universal_tag(t))
                 for (w, t) in sent]
                for sent in nltk.corpus.floresta.tagged_sents()]
    shuffle(floresta)

    mac_morpho = [[w[0] for w in sent] for sent in
                  nltk.corpus.mac_morpho.tagged_paras()]
    mac_morpho = [
        [(w, convert_to_universal_tag(t, reverse=True))
         for (w, t) in sent] for sent in mac_morpho]
    shuffle(mac_morpho)

    regex_patterns = [
        (r"^[nN][ao]s?$", "ADP"),
        (r"^[dD][ao]s?$", "ADP"),
        (r"^[pP]el[ao]s?$", "ADP"),
        (r"^[nN]est[ae]s?$", "ADP"),
        (r"^[nN]um$", "ADP"),
        (r"^[nN]ess[ae]s?$", "ADP"),
        (r"^[nN]aquel[ae]s?$", "ADP"),
        (r"^\xe0$", "ADP"),
    ]

    def_tagger = nltk.DefaultTagger('NOUN')
    affix_tagger = nltk.AffixTagger(
        mac_morpho + floresta, backoff=def_tagger
    )
    unitagger = nltk.UnigramTagger(
        mac_morpho + floresta, backoff=affix_tagger
    )
    rx_tagger = nltk.RegexpTagger(
        regex_patterns, backoff=unitagger
    )
    tagger = nltk.BigramTagger(
        floresta, backoff=rx_tagger
    )
    tagger = nltk.BrillTaggerTrainer(tagger, nltk.brill.fntbl37())
    tagger = tagger.train(floresta, max_rules=100)

    with open(path, "wb") as f:
        pickle.dump(tagger, f)

    return tagger


def load_pt_tagger(path=None):
    path = path or join(dirname(dirname(__file__)), "res", "pt_tagger.pkl")
    if not isfile(path):
        train_pt_tagger(path)
    with open(path, "rb") as f:
        tagger = pickle.load(f)
    return tagger


_POSTAGGER = load_pt_tagger()


def pos_tag_pt(tokens):
    if isinstance(tokens, str):
        tokens = word_tokenize(tokens)
    postagged = _POSTAGGER.tag(tokens)

    # HACK this fixes some know failures from postag
    # this is not sustainable but important cases can be added at any time
    # PRs + unittests welcome!
    DETS = ["a", "á", "o", "ós", "aos", "ao"]
    for idx, (w, t) in enumerate(postagged):
        next_w, next_t = postagged[idx + 1] if \
            idx < len(postagged) - 1 else ("", "")

        #  ('á', 'NOUN'), ('Maria', 'NOUN')
        if w.lower() in DETS and t == "NOUN":
            postagged[idx] = (w, "DET")

    return postagged


# word rules for gender
_FEMALE_ENDINGS_PT = ["a", "as"]
_MALE_ENDINGS_PT = ["o", "os"]


def predict_gender_pt(word, text=""):
    # parse gender taking context into account
    word = word.lower()
    words = text.lower().split(" ")
    for idx, w in enumerate(words):
        if w == word and idx != 0:
            # in portuguese usually the previous word (a determinant)
            # assigns gender to the next word
            previous = words[idx - 1].lower()
            if previous in MALE_DETERMINANTS_PT:
                return "male"
            elif previous in FEMALE_DETERMINANTS_PT:
                return "female"

    # get gender using only the individual word
    # see if this word has the gender defined
    if word in GENDERED_WORDS_PT["male"]:
        return "male"
    if word in GENDERED_WORDS_PT["female"]:
        return "female"
    singular = word.rstrip("s")
    if singular in GENDERED_WORDS_PT["male"]:
        return "male"
    if singular in GENDERED_WORDS_PT["male"]:
        return "female"
    # in portuguese the last vowel usually defines the gender of a word
    # the gender of the determinant takes precedence over this rule
    for end_str in _FEMALE_ENDINGS_PT:
        if word.endswith(end_str):
            return "female"
    for end_str in _MALE_ENDINGS_PT:
        if word.endswith(end_str):
            return "male"
    return None
